# pylint: disable=undefined-variable
# pylint: disable=import-error

"""
Module that contains classes representing love letter's cards
"""

import gettext

from telebot import types

gettext.install('loveletter', localedir='./loveletter/locale', codeset='UTF-8')


class Card:
    """
    An abstract class that represents the arbitrary card

    :static attr name:
        str, name of the card
    :static attr value:
        int, value of the card, the better card allows you to win in the end
    :static attr targeted:
        bool, checks if card needs to point another player as a target
    :static attr num_in_deck:
        int, number of cards of this type in standard deck
    :attr owner:
        User, player who holds this card right now
    """

    name = None
    value = None
    targeted = None
    num_in_deck = None

    def __init__(self):
        self.owner = None

    def play(self, game):
        """
        Virtual function, that plays this card
        within given game and applies feature of specific card

        :param game:
            Game, game in which this card is played
        """

        raise NotImplementedError

    def __eq__(self, other):
        return self.name == other

    def __le__(self, other):
        return self.value <= other.value

    def __ge__(self, other):
        return self.value >= other.value

    def __lt__(self, other):
        return self.value < other.value

    def __gt__(self, other):
        return self.value > other.value


class Princess(Card):
    """
    Princess card

    The Princess is most powerful card because player who holds it
    at the end of the game definitely wins. But if player accidentally
    drop this card he would instantly loose. As soon as playing this card
    assumes, you should drop it from hand, playing it is surely a suicide)
    """
    # pylint: disable=too-few-public-methods

    name = _("Princess")
    value = 8
    targeted = False
    num_in_deck = 1

    def play(self, game):
        markup = types.ReplyKeyboardRemove(selective=False)
        message = _("@{} drops a Princess and loses.").format(self.owner.name)

        game.public_message(message, markup)
        game.used_cards.append(self.owner.card)
        game.users.kill(self.owner)


class Countess(Card):
    """
    Countess card

    Has the less power than Princess, but still it better than other cards
    Just drop it you wanna play it. Yes it does absolutely nothing, if you don't
    count that it must be played if you have King or Prince cards (and actually Princess too)
    """
    # pylint: disable=too-few-public-methods

    name = _("Countess")
    value = 7
    targeted = False
    num_in_deck = 1

    def play(self, game):
        markup = types.ReplyKeyboardRemove(selective=False)
        message = _("@{} drops a Countess.").format(self.owner.name)

        game.public_message(message, markup)


class King(Card):
    """
    The King card

    King has less value than Countess but greater than Prince's one
    Playing the King means you drop him from hand and change your second card with
    selected player's one
    """
    # pylint: disable=too-few-public-methods

    name = _("King")
    value = 6
    targeted = True
    num_in_deck = 1

    def play(self, game):
        markup = types.ReplyKeyboardRemove(selective=False)

        if game.card_without_action:
            message = \
                _("@{} drops the King because all players are protected.").format(self.owner.name)

            game.public_message(message, markup)

            return

        message = \
            _("@{0} plays the King to exchange cards with @{1}.").format(self.owner.name,
                                                                         game.victim.name)

        game.public_message(message, markup)

        message = _("You've got a {0} from player @{1}").format(game.victim.card.name,
                                                                game.victim.name)
        game.dealer_message(message)

        message = _("You've got a {0} from player @{1}").format(game.dealer.card.name,
                                                                game.dealer.name)
        game.victim_message(message)

        game.dealer.card.owner = game.victim
        game.victim.card.owner = game.dealer

        game.dealer.card, game.victim.card = game.victim.card, game.dealer.card


class Prince(Card):
    """
    The Prince card

    Prince stands one step over Maid but one step below the King
    This card allows you to make one other player to drop off his card
    (note it: if he has the Princess, he will loose), and take a new one from deck.
    """
    # pylint: disable=too-few-public-methods

    name = _("Prince")
    value = 5
    targeted = True
    num_in_deck = 2

    def play(self, game):
        markup = types.ReplyKeyboardRemove(selective=False)
        game.used_cards.append(game.victim.card)

        if isinstance(game.victim.card, Princess):
            message = _("@{0} uses Prince against @{1}. "
                        "@{1} drops a Princess and loses!").format(self.owner.name,
                                                                   game.victim.name)

            game.public_message(message, markup)
            game.users.kill(game.victim)

            return

        message = \
            _("@{0} uses a Prince against @{1}. @{1} has the {2}.").format(self.owner.name,
                                                                           game.victim.name,
                                                                           game.victim.card.name)

        game.public_message(message, markup)

        if not game.deck:
            game.deck.append(game.first_card)

        game.victim.take_card(game.deck)


class Maid(Card):
    """
    The Maid card

    The fourth card in the hierarchy
    Using this card makes you protected from actions of
    any targeted cards for the entire round (till your next turn)
    """
    # pylint: disable=too-few-public-methods

    name = _("Maid")
    value = 4
    targeted = False
    num_in_deck = 2

    def play(self, game):
        markup = types.ReplyKeyboardRemove(selective=False)
        message = _("@{} is under Maid protection for a one full round.").format(self.owner.name)

        game.public_message(message, markup)

        self.owner.defence = True


class Baron(Card):
    """
    The Baron card

    The third card by it's power
    Playing it allowes you to take out any player, but only if
    your second card (that remains in your hand) is greater than card
    which this player holds, else you will take out yourself.
    """
    # pylint: disable=too-few-public-methods

    name = _("Baron")
    value = 3
    targeted = True
    num_in_deck = 2

    def play(self, game):
        markup = types.ReplyKeyboardRemove(selective=False)

        if game.card_without_action:
            message = \
                _("@{} drops a Baron, because all players are protected.").format(self.owner.name)

            game.public_message(message, markup)

            return

        if self.owner.card > game.victim.card:
            message = _("@{0} uses the Baron against @{1} and wins. "
                        "@{1} has the {2} card").format(self.owner.name, game.victim.name,
                                                        game.victim.card.name)

            game.public_message(message, markup)

            game.used_cards.append(game.victim.card)
            game.users.kill(game.victim)

            return

        if self.owner.card < game.victim.card:
            message = _("@{0} uses the Baron against @{1}, but looses. "
                        "@{0} has the {2} card").format(self.owner.name, game.victim.name,
                                                        self.owner.card.name)

            game.public_message(message, markup)

            game.used_cards.append(self.owner.card)
            game.users.kill(self.owner)

            return

        message = \
            _("Aaaand... here goes nothing"
              "Looks like @{0} and @{1} have the same cards... ").format(self.owner.name,
                                                                         game.victim.name)

        game.public_message(message, markup)


class Priest(Card):
    """
    The Priest card

    Not very strong card (actually it can beat only Guard)
    but it allows you to peek anyone else's card, you just need to play it
    """
    # pylint: disable=too-few-public-methods

    name = _("Priest")
    value = 2
    targeted = True
    num_in_deck = 2

    def play(self, game):
        markup = types.ReplyKeyboardRemove(selective=False)

        if game.card_without_action:
            message = \
                _("@{} drops Priest, because all players are protected.").format(self.owner.name)
            game.public_message(message, markup)

            return

        message = _("@{} uses a Priest card and looks at @{}'s card.").format(self.owner.name,
                                                                              game.victim.name)
        game.public_message(message, markup)

        message = _("@{} shows you his card. He(She) has the {}.").format(game.victim.name,
                                                                          game.victim.card.name)
        game.dealer_message(message)


class Guard(Card):
    """
    The Guard card

    The weakest card in the game, if you finish game with it
    you definitely won't win (only if you don't kicked out other players),
    but it allows you to kick one player from game, you just need to play it
    and guess which card this player holds.
    """
    # pylint: disable=too-few-public-methods

    name = _("Guard")
    value = 1
    targeted = True
    num_in_deck = 5

    def play(self, game):
        markup = types.ReplyKeyboardRemove(selective=False)

        if game.card_without_action:
            message = \
                _("@{} drops the Guard, because all players are protected.").format(self.owner.name)
            game.public_message(message, markup)

            return

        if game.victim.card.name == game.guess:
            message = \
                _("@{} uses the Guard and guess that @{} has the {}.").format(self.owner.name,
                                                                              game.victim.name,
                                                                              game.guess)

            game.public_message(message, markup)

            game.used_cards.append(game.victim.card)
            game.users.kill(game.victim)

            return

        message = \
            _("@{0} uses Guard, assuming that "
              "@{1} holds the {2} card, but do not guess it right.").format(self.owner.name,
                                                                            game.victim.name,
                                                                            game.guess)
        game.public_message(message, markup)



# pylint: disable=undefined-variable
# pylint: disable=import-error

"""
Here contains the Game class that represents
the game table and handles all events within one game
"""

import random
from itertools import chain
import gettext

from telebot import types

from loveletter.users import Users
from loveletter.cards import (
    Princess,
    Countess,
    King,
    Prince,
    Maid,
    Baron,
    Priest,
    Guard
)

gettext.install('loveletter', localedir='./loveletter/locale', codeset='UTF-8')


class Game:
    """
    Class that handles one game, a virtual table if you prefer

    Game has several states, running cycle is:

    not_started (initial state) ─→ change_turn
    change_turn ─→ select_card
    select_card ┬→ select_victim (if card is targeted)
                ├→ change_turn
                └→ not_started (if game over)
    select_victim ┬→ guess_card (if card is Guard)
                  ├→ change_turn
                  └→ not_started (if game over)
    guess_card ┬→ change_turn
               └→ not_started (if game over)
    """

    card_types = (
        Princess,
        Countess,
        King,
        Prince,
        Maid,
        Baron,
        Priest,
        Guard,
    )

    def __init__(self, bot):
        """
        Creates a new game

        TODO This class has too many instances and needs to be broken into more abstract subclasses

        :param bot:
            Bot, the bot that handles all the player-to-game interactions
        :param game_id:
            int, unique game id (also user_id of the game creator)
        """

        self.bot = bot
        self.users = Users()
        self.used_cards = []
        self.dealer = None
        self.victim = None
        self.guess = None
        self.first_card = None
        self.can_choose_yourself = False
        self.card_without_action = False
        self.double_deck = False
        self.state = 'not_started'

        self.deck = self.generate_deck()

    def start(self):
        """
        Starts a new game and deals the cards.
        This method must be called after adding
        all players to game and configuring it.

        current_state: not_started
        next_state: change_turn
        """

        if self.state != 'not_started':
            raise RuntimeError('Trying to start a game, when it is already started')

        if self.double_deck:
            self.deck.extend(self.generate_deck())

        random.shuffle(self.deck)
        self.first_card = self.deck.pop()

        self.users.shuffle()
        for user in self.users:
            user.take_card(self.deck)

        message = _("The game is started!\n"
                    "Players order (top moves first):\n")

        for num, user in enumerate(self.users):
            if num == 0:
                message += '\t@{} <<\n'.format(user.name)
            else:
                message += '\t@{}\n'.format(user.name)

        self.public_message(message)

        self.state = 'change_turn'

    def restart(self):
        """
        Restarts the game

        current state: any
        next state: not_started
        """
        self.users.reset()
        self.deck = self.generate_deck()
        self.used_cards = []

        self.state = 'not_started'

        self.start()
        self.start_turn()

    def is_game_over(self):
        """
        Checks if game is over

        game ends if it is only one player left,
        or there is no cards in deck.
        If game is ended, sends game result to chat.
        If game is not ended, starts a new turn

        current state: any
        next state:
            game_over (if game ended)
            change_turn (if game not ended)
        """

        if len(self.users) == 1 or not self.deck:
            players_remains = sorted(self.users, key=lambda user: user.card, reverse=True)
            winner = players_remains[0]

            message = [_("Game is over\n"
                         "Winner is {}\n"
                         "Players remains\n").format(winner.name)]

            for i, user in enumerate(players_remains):
                message.append("\t#{} @{} - {} ({})\n".format(i+1, user.name, user.card.name,
                                                              user.card.value))

            message.append(_("Kicked off the game:\n"))

            for i, user in enumerate(self.users.loosers):
                message.append("\t@{}\n".format(user.name))

            self.public_message(''.join(message))

            self.bot.send_message(winner.user_id, _("Greetings! You've won!"))
            for user in self.users:
                if user != winner:
                    self.bot.send_message(user.user_id, _("You loose!"))

            self.state = 'game_over'

            return

        self.state = 'change_turn'
        self.start_turn()

    def start_turn(self):
        """
        Starts a player's turn

        If player has Maid's defence, removes it,
        sends messages to chat about number of card left and
        current player, make current dealer take new card
        offers player to choose which card to play.

        current state: change_turn
        next state: select_card
        """

        if self.state != 'change_turn':
            raise RuntimeError('Trying to start a turn while not in change_turn state')

        self.dealer = self.users.get_dealer()
        self.dealer.defence = False

        self.public_message(_("@{}'s turn").format(self.dealer.name))
        self.dealer.take_new_card(self.deck)

        if self.deck:
            self.public_message(_("It's {} cards left.").format(len(self.deck)))
        else:
            self.public_message(_("Attention! It's the last turn"))

        markup = types.ReplyKeyboardMarkup(row_width=2)
        button1 = types.KeyboardButton(self.dealer.card.name)
        button2 = types.KeyboardButton(self.dealer.new_card.name)
        markup.add(button1, button2)

        self.dealer_message(_("Choose a card which you want to play:"), markup)

        self.state = 'select_card'

    def select_card(self, card_name):
        """
        Applies card features selected by user in game

        If card is targeted, game changes state to select a victim,

        :param card_name:
            str, name of the card, which player want to play

        current_state: select_card
        next_state:
            select_victim (if card is targeted)
            change_turn (else)
        """
        if self.state != 'select_card':
            raise RuntimeError('Trying to select card while not in select_card state')

        if card_name in [_("Prince"), _("King")] and (isinstance(self.dealer.card, Countess) or
                                                      isinstance(self.dealer.new_card, Countess)):
            self.dealer_message(_("Woopsy-daisy... You need to drop a countess."))
            return

        if card_name != self.dealer.new_card.name:
            self.dealer.new_card, self.dealer.card = self.dealer.card, self.dealer.new_card

        if self.dealer.new_card.targeted:
            self.state = 'select_victim'

            self.can_choose_yourself = False
            self.card_without_action = False

            markup = types.ReplyKeyboardMarkup()
            possible_victims = self.list_possible_victims()

            for user_name in possible_victims:
                button = types.KeyboardButton(user_name)
                markup.add(button)

            if not self.card_without_action:
                self.dealer_message(_("Choose the player you want play this card with:"), markup)
                return

        active_card = self.dealer.new_card

        self.dealer.new_card = None
        self.used_cards.append(active_card)

        active_card.play(self)

        self.state = 'change_turn'
        self.is_game_over()

    def select_victim(self, victim_name):
        """

        :param victim_name:
        :return:
        """

        if self.state != 'select_victim':
            raise RuntimeError("Trying to select a victim, while not in select_victim state")

        self.victim = self.users.find_by_name(victim_name)

        if self.card_without_action:
            active_card = self.dealer.new_card
            self.dealer.new_card = None
            self.used_cards.append(active_card)
            active_card.play(self)
            self.is_game_over()

            self.state = 'change_turn'

            return

        if isinstance(self.dealer.new_card, Guard):
            markup = types.ReplyKeyboardMarkup()

            for card in Game.card_types[:-1]:
                button = types.KeyboardButton(card.name)
                markup.add(button)

            self.dealer_message(_("Guess the @{}'s card:").format(self.victim.name), markup)
            self.state = 'guess_card'
            return

        self.state = 'change_turn'

        active_card = self.dealer.new_card
        self.dealer.new_card = None
        self.used_cards.append(active_card)

        active_card.play(self)
        self.is_game_over()

    def guess_card(self, guess):
        """

        :param guess:
        :return:
        """

        if self.state != 'guess_card':
            raise RuntimeError('Guessing card while not in guess state')

        active_card = self.dealer.new_card
        self.dealer.new_card = None
        self.used_cards.append(active_card)

        self.guess = guess
        active_card.play(self)

        self.is_game_over()

    def list_possible_victims(self):
        """
        Returns a list of possible victims within one game
        Player can be a victim only if he is not defended by Maid
        and he is not a dealer. Howewer if everybody are defended,
        dealer can become the only one victim, or he can use Prince on himself

        :return:
            list, list of victims
        """
        victims_list = []
        for user in self.users:
            if not user.defence and user != self.dealer:
                victims_list.append(user.name)

        if isinstance(self.dealer.new_card, Prince) or len(victims_list) == 0:
            victims_list.append(self.dealer.name)

            if not isinstance(self.dealer.new_card, Prince):
                self.card_without_action = True
            else:
                self.can_choose_yourself = True

        return victims_list

    def public_message(self, message, markup=None, but=None):
        """
        sends message to all users in this game

        :param message:
            message to be sent
        :param markup:
            markup with helper buttons
        :param but:
            int, user_id, who don't need a message
        """

        for user in self.users:
            if but is not None and user == but:
                continue
            self.bot.send_message(user.user_id, message, reply_markup=markup)

        for user in self.users.loosers:
            if but is not None and user == but:
                continue
            self.bot.send_message(user.user_id, message, reply_markup=markup)

    def dealer_message(self, message, markup=None):
        """
        sends message to dealer

        :param message:
            message to be sent
        :param markup:
            markup with helper buttons
        """
        self.bot.send_message(self.dealer.user_id, message, reply_markup=markup)

    def victim_message(self, message, markup=None):
        """
        sends message to victim

        :param message:
           message to be sent
        :param markup:
           markup with helper buttons
        """
        self.bot.send_message(self.victim.user_id, message, reply_markup=markup)

    @staticmethod
    def generate_deck():
        """
        Generates a standard deck of cards

        :return:
            list of Cards, the deck
        """
        return list(chain(*[
            [Card() for _ in range(Card.num_in_deck)]
            for Card in Game.card_types
        ]))



# pylint: disable=undefined-variable
# pylint: disable=import-error

"""
Module contains classes that represents a players
User and Users (see their docstrings for more)
"""

import random
from collections import deque
import gettext

gettext.install('loveletter', localedir='./loveletter/locale', codeset='UTF-8')


class User:
    """
    class that represents a player

    :attr name:
        user's @nickname from telegram
    :attr user_id:
        inner telegram id, used for writing
        private messages to user
    :attr bot:
        link to the bot which handles game where
        this user is playing
    :attr card:
        currently card on the user's hand
    :attr new_card:
        card that was taken to hand on the move
    :attr defence:
        is this user protected by 'Maid' card
    """

    def __init__(self, name, user_id, bot, game):
        """
        Creates a new user
        As soon as user can be created only
        before the game starts, all game attrs (such as card,
        new_card and defence) are set to None

        :param name:
            user's @nickname from telegram
        :param user_id:
            inner telegram id, used for writing
            private messages to user
        :param bot:
            link to the bot which handles game where
            this user is playing
        """

        self.name = name
        self.user_id = user_id
        self.bot = bot
        self.game = game
        self.card = None
        self.new_card = None
        self.defence = False

    def take_card(self, deck):
        """
        Takes a top card from the deck,
        if user currently not holding any cards
        (at start of the game or after he drops all his cards)

        :param deck:
            list of Cards, deck from where the card is taken
        """

        self.card = deck[-1]
        self.card.owner = self
        del deck[-1]
        self.bot.send_message(self.user_id, _("Your card is '{}'").format(self.card.name))

    def take_new_card(self, deck):
        """
        Take new card on user's turn, so he handle
        two cards at the same time. One of the cards must be played
        on this turn

        :param deck:
            list of Cards, deck from where the card is taken
        """

        self.new_card = deck[-1]
        self.new_card.owner = self
        del deck[-1]
        self.bot.send_message(self.user_id,
                              _("You have taken the '{}' card").format(self.new_card.name))

    def __eq__(self, other):
        if isinstance(other, User):
            return self.user_id == other.user_id
        return self.user_id == other

    def __ne__(self, other):
        return not self == other


class Users:
    """
    Class that represents a users queue
    Actually it just a wrapper over the deque,
    but provides several useful methods

    :attr queue:
        deque of Users, users who currently playing the game
        and do not lost yet
    :attr loosers:
        players who was kicked from the game
    """

    def __init__(self):
        """
        Creates empty users queue
        """

        self.queue = deque()
        self.loosers = []

    def add(self, user):
        """
        Adds user to the users queue

        :param user:
            user to be added
        """
        self.queue.append(user)

    def shuffle(self):
        """
        Randomly shuffles all users in queue
        """
        random.shuffle(self.queue)

    def get_dealer(self):
        """
        Select next user (aka dealer) from queue and
        moves it to the end of queue
        """

        user = self.queue.popleft()
        self.queue.append(user)
        return user

    def kill(self, user):
        """
        Throw given user out of the game

        :param user:
            User, who loose the game
        """

        user.bot.send_message(user.user_id, _("You've lost!"))
        self.queue.remove(user)
        self.loosers.append(user)

    def get_victims(self, dealer):
        """
        Get users that could be possible victims for
        targeted cards, taking into account that dealer
        cannot be victim of himself

        :param dealer:
            User, who totally not a victim
        """

        victims = []
        for user in self.queue:
            if not user.defence and user != dealer:
                victims.append(user.name)
        return victims

    def num_users(self):
        """
        Get number of players,
        which didn't loose at this moment

        :return:
            int, number of players
        """

        return len(self.queue)

    def find_by_name(self, name):
        """
        Finds user in queue by name

        :param name:
            A name to find user by
        """
        for user in self.queue:
            if user.name == name:
                return user

        return None

    def reset(self):
        """
        Resets all lost users back to queue
        """

        self.queue.extend(self.loosers)
        self.loosers = []

        self.shuffle()

    def __contains__(self, user):
        return user in self.queue

    def __iter__(self):
        return iter(self.queue)

    def __len__(self):
        return len(self.queue)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name="loveletter",
      version="0.0.5",
      author="Konstantin Kozlovtsev",
      packages=find_packages(),
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/kst179/LoveLetter",
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
      ],
      entry_points={
          'console_scripts': [
                'loveletter = loveletter.bot:main'
          ],
      }
      )



import logging
import sys
import time
import random

import unittest
from telebot import types

from loveletter.bot import GameBot
from loveletter.cards import Guard
from loveletter.game import Game


class TestBot(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.users_id = 0
        self.messages_id = 0

        logging.basicConfig(
            stream=sys.stderr,
            level=logging.INFO,
            format='[%(levelname)s:%(asctime)s %(filename)s:%(lineno)s] %(message)s',
            datefmt='%Y-%m-%d %I:%M:%S'
        )

    def test_create_joins(self):
        bot = self.setup_bot()
        alice = self.new_user('alice')
        bob = self.new_user('bob')
        cinderella = self.new_user('cinderella')

        bot.process_new_messages([
            self.new_message(alice, '/create'),

            # wrong joins
            self.new_message(bob, '/join'),
            self.new_message(bob, '/join @cinderella'),

            # right join
            self.new_message(bob, '/join @alice'),

            # check if they cannot join after join
            self.new_message(alice, '/join @bob'),
            self.new_message(bob, '/join @alice'),

            # check if nobody cannot join after start
            self.new_message(alice, '/start'),
            self.new_message(cinderella, '/join @alice'),
            self.new_message(cinderella, '/join @bob'),
        ])

        # bot is asynchronous so we need to wait messages processing
        time.sleep(1)

        game = bot.get_game(alice.id)

        self.assertTrue(alice.id in game.users)
        self.assertTrue(bob.id in game.users)
        self.assertFalse(cinderella.id in game.users)

    def test_play_random_game(self):
        bot = self.setup_bot()
        users, game = self.setup_game_with_4_players(bot)

        self.play_random_game(bot, users, game)

        self.assertTrue(len(game.users) == 1 or not game.deck)

    def test_restart_game(self):
        bot = self.setup_bot()
        users, game = self.setup_game_with_4_players(bot)

        self.play_random_game(bot, users, game)

        bot.process_new_messages([
            self.new_message(users[0], '/restart')
        ])

        time.sleep(0.1)

        self.assertNotEqual(game.state, 'game_over')
        self.assertTrue(not game.users.loosers)

        self.play_random_game(bot, users, game)

    def setup_game_with_4_players(self, bot):
        users = [
            self.new_user('alice'),
            self.new_user('bob'),
            self.new_user('cinderella'),
            self.new_user('dolly'),
        ]

        bot.process_new_messages([
            self.new_message(users[0], '/create'),
            self.new_message(users[1], '/join @alice'),
            self.new_message(users[2], '/join @alice'),
            self.new_message(users[3], '/join @alice'),
            self.new_message(users[0], '/start')
        ])

        time.sleep(1)

        game = bot.get_game(users[0].id)

        return users, game

    def play_random_game(self, bot,  users, game):
        while game.state != 'game_over':
            player = game.dealer

            for user in users:
                if user.id == player:
                    break

            card = random.choice([player.card, player.new_card])

            bot.process_new_messages([
                self.new_message(user, card.name)
            ])

            time.sleep(0.1)

            if not card.targeted or game.card_without_action:
                continue

            victim = random.choice(game.list_possible_victims())

            bot.process_new_messages([
                self.new_message(user, victim)
            ])

            time.sleep(0.1)

            if not isinstance(card, Guard):
                continue

            guess = random.choice([card.name for card in Game.card_types[:-1]])

            bot.process_new_messages([
                self.new_message(user, guess)
            ])

        return game

    def new_user(self, name):
        user_id = self.users_id
        self.users_id += 1

        user = types.User(
            id=user_id,
            is_bot=False,
            first_name=name,
            username=name,
        )

        return user

    def new_message(self, user: types.User, message):
        message_id = self.messages_id
        self.messages_id += 1

        chat = types.Chat(
            id=user.id,
            type='private'
        )

        message = types.Message(
            message_id=message_id,
            from_user=user,
            date=None,
            chat=chat,
            content_type='text',
            options={'text': message},
            json_string='',
        )

        return message

    @staticmethod
    def setup_bot():
        def fake_send_message(chat_id, text, reply_markup=None, parse_mode=None):
            logging.info("[#%d] %s", chat_id, text)

        bot = GameBot('')
        bot.send_message = fake_send_message

        return bot


if __name__ == '__main__':
    tester = TestBot()
    tester.run()



# pylint: disable=undefined-variable
# pylint: disable=import-error

"""
module docstring
"""
import gettext
import logging
import argparse
import os
import sys

import telebot

import numpy as np

from loveletter.game import Game
from loveletter.users import User

gettext.install('loveletter', localedir='./loveletter/locale', codeset='UTF-8')


class GameBot(telebot.TeleBot):
    """
    Class that represents a TelegramBot
    that handles all games and players interactions

    :attribute games: dict of {chat_id: game}
    """

    def __init__(self, token):
        """
        Creates a bot

        :param token:
            token (str) for game bot, if you don't have one,
            you should go to @BotFather in telegram and
            create a new bot (just send him a '/newbot'
            message and follow instructions, it's pretty easy)
        """
        super().__init__(token)
        self.games = {}
        self.users = {}
        self.name2user = {}
        self.register_handlers()

    def register_handlers(self):
        """
        Registers handlers for bot.
        Handler is a function with '@bot.message_handler' descriptor,
        once created, bot will continuously accept special commands from players
        and handle it.
        """
        # This functions would be called only by base class methods,
        # so linter thinks they are unused
        # pylint: disable=unused-variable

        @self.message_handler(commands=['help'])
        def show_help(message):
            self.show_help(message)

        @self.message_handler(commands=['hint'])
        def show_hint(message):
            self.show_hint(message)

        @self.message_handler(commands=['create'])
        def create_game(message):
            self.create_game(message)

        @self.message_handler(commands=['doubledeck'])
        def double_deck(message):
            self.double_deck(message)

        @self.message_handler(commands=['join'])
        def join_user(message):
            self.join_user(message)

        @self.message_handler(commands=['leave'])
        def leave_game(message):
            self.leave_game(message)

        @self.message_handler(commands=['start'])
        def start_game(message):
            self.start_game(message)

        @self.message_handler(commands=['restart'])
        def restart_game(message):
            self.restart_game(message)

        @self.message_handler(commands=['cards'])
        def show_cards(message):
            self.show_cards(message)

        @self.message_handler(commands=['players'])
        def show_users(message):
            self.show_users(message)

        @self.message_handler(content_types=['text'])
        def text_handler(message):
            self.text_handler(message)

    def get_game(self, user_id):
        """
        Helper function to get game by chat_id
        where it is playing, if game not found returns None

        :param user_id:
            int, is of one of the players
        :return:
            Game, if game was created in chat with given id,
            else None
        """
        if user_id not in self.users.keys():
            self.send_message(
                user_id,
                _("You didn't join to any game yet, try send me a /create or /join <game_id>")
            )

            return None

        game = self.users[user_id].game

        return game

    def show_help(self, message):
        """
        Shows help in chat

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        """

        self.send_message(message.chat.id, _("help"))

    def show_hint(self, message):
        """
        Shows hint (small copy of game rules) in chat

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        """
        self.send_message(message.chat.id, _("hints"))

    def create_game(self, message):
        """
        Creates game in chat from where message is came

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        """
        user_id = message.from_user.id
        user_name = message.from_user.username or message.from_user.username

        if user_id in self.users.keys():
            self.send_message(user_id,
                              _("The game has been already created in this chat, restarting it"))

        game = Game(self)
        user = User(user_name, user_id, self, game)

        game.users.add(user)
        self.users[user_id] = user
        self.name2user[user_name] = user

        self.send_message(user_id, _("Game is created, resend next message to your "
                                     "freinds whith whom you would like to play").format(user_id))

        print(self.get_me().username)
        self.send_message(user_id,
                          _("Player @{0} invites you to love letter game, "
                            "send message <code>/join @{0}</code> here @{1} "
                            "if you want accept it").format(user_name, self.get_me().username),
                          parse_mode='HTML')

        logging.info('Chat #%d: game created', user_id)

    def double_deck(self, message):
        """
        Toggles second set of cards,
        it can be useful if more than 5
        players are in game

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        """
        user_id = message.user.id
        username = message.user.name or message.from_user.username

        game = self.get_game(user_id)

        if not game:
            return

        if game.state != 'not_started':
            self.send_message(user_id, _("The game is already started"))
            return

        if not game.double_deck:
            game.double_deck = True
            game.public_message(_("The second deck is added by @{}".format(username)))
        else:
            game.double_deck = False
            game.public_message(_("The second deck is removed by @{}".format(username)))

        logging.info('Chat #%d: set doubledeck', user_id)

    def join_user(self, message):
        """
        Adds user who wrote this message to game

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        """
        text = message.text.split()
        user_id = message.from_user.id
        username = message.from_user.username or message.from_user.first_name

        logging.info(text)

        if len(text) < 2 or not text[1].startswith('@'):
            self.send_message(user_id, _("There is no username, try /join @username or /create"))
            return

        friend_name = text[1][1:]

        if friend_name not in self.name2user.keys():
            self.send_message(user_id,
                              _("There is no game with such player @{}").format(friend_name))
            return

        friend = self.name2user[friend_name]

        game = self.get_game(friend.user_id)

        if not game:
            return

        if game.state != 'not_started':
            self.send_message(user_id, _("Game has been already started"))
            return

        if user_id in game.users:
            self.send_message(user_id, _("You already joined to game"))
            return

        user = User(username, user_id, self, game)
        self.users[user_id] = user
        self.name2user[username] = user

        game.users.add(user)
        game.public_message('Player @{} joined to game'.format(username))

        if game.users.num_users() == 6 and not game.double_deck:
            game.double_deck = True
            game.public_message(
                _("The number of players reached 6, the second deck is automatically added")
            )

        logging.info('Chat #%d: user #%d added', friend.user_id, user_id)

    def leave_game(self, message):
        user_id = message.from_user.id

        game = self.get_game(user_id)

        if game is None:
            return

        user = self.users[user_id]

        if user in game.users:
            if game.state != 'not_started':
                self.send_message(user_id, _("The game already started, you cannot leave!"))
                return

            game.users.queue.remove(user)

            self.users.pop(user_id)
            self.name2user.pop(user.name)

            self.send_message(user_id, _("You left the game"))
            return

        if user in game.users.loosers:
            game.users.loosers.remove(user)

            self.users.pop(user_id)
            self.name2user.pop(user.name)

            self.send_message(user_id, _("You left the game"))

    def start_game(self, message):
        """
        Starts game, and make first turn.
        There is no way to change number of players or
        toggle double deck if game is already started.

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        """
        user_id = message.from_user.id

        game = self.get_game(user_id)

        if not game:
            return

        if len(game.users) < 2:
            self.send_message(user_id,
                              _("Not enough players, to play, you need at least 2 of them"))
            return

        game.start()
        game.start_turn()

        logging.info("Chat #%d: game started", user_id)

    def restart_game(self, message):
        """
        Restarts game

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        """

        user_id = message.from_user.id

        game = self.get_game(user_id)

        if not game:
            return

        game.restart()

        logging.info("Chat #%d: game restarted", user_id)

    def show_cards(self, message):
        """
        Shows all cards that was already played

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        """
        user_id = message.from_user.id
        game = self.get_game(user_id)

        if game is None:
            return

        used_cards = _("Dropped cards list:\n")
        unique_used_cards = np.unique(sorted(game.used_cards), return_counts=True)

        for num, card in enumerate(unique_used_cards[0]):
            used_cards += ' - {:10s} [{}]\n'.format(card.name, unique_used_cards[1][num])

        self.send_message(user_id, used_cards)

    def show_users(self, message):
        """
        Shows all players, who is current dealer,
        and who is currently protected by the 'Maid' card to chat

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        :return:
        """
        user_id = message.from_user.id
        game = self.get_game(user_id)

        if not game:
            return

        users = _("Players remained: \n")
        for user in game.users:
            users += ' - @{} '.format(user.name)
            if user.defence:
                users += '^'
            if user == game.dealer:
                users += '<<'
            users += '\n'

        self.send_message(user_id, users)

    def text_handler(self, message):
        """
        Handles all free-text messages, that are
        used for moving between states

        :param message:
            telebot.types.Message, message that contains
            info about chat where it was written and user who wrote it
        """
        user_id = message.from_user.id
        user_name = message.from_user.username
        game = self.get_game(user_id)

        if not game:
            return

        if game.state != 'not_started' and user_id == game.dealer.user_id:
            if game.state == 'select_card' and \
                    message.text in [game.dealer.card.name, game.dealer.new_card.name]:
                game.select_card(message.text)
                return

            if (game.state == 'select_victim' and
                    (message.text in game.users.get_victims(game.dealer) or
                     game.can_choose_yourself and message.text == game.dealer.name)):
                game.select_victim(message.text)
                return

            if game.state == 'guess_card':
                game.guess_card(message.text)
                return

        game.public_message('@{}: {}'.format(user_name, message.text), but=user_id)


def main():
    """
    Starts a bot

    To correct work you need to specify bot token in --token arg
    or in os environment variable LOVELETTER_TOKEN
    """

    logging.basicConfig(
        stream=sys.stdout,
        level=logging.INFO,
        format='[%(levelname)s:%(asctime)s %(filename)s:%(lineno)s] %(message)s',
        datefmt='%Y-%m-%d %I:%M:%S'
    )

    parser = argparse.ArgumentParser()
    parser.add_argument('--token', type=str, default=None)

    args = parser.parse_args()
    token = args.token

    if token is None:
        if 'LOVELETTER_TOKEN' not in os.environ:
            raise RuntimeError('No token specified')

        token = os.environ['LOVELETTER_TOKEN']

    logging.info("Bot started")

    bot = GameBot(token)
    bot.polling(none_stop=True)


if __name__ == '__main__':
    main()
